/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2006,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#ifndef _COS_SIN14_H_
#define _COS_SIN14_H_	1

#ifndef M_1_2PI
#define M_1_2PI		0.1591549430919f	/* 1/(2*pi) */ 
#endif /* M_1_2PI */

/*
 * FUNCTION
 *	float _cos_sin14(float angle)
 *
 * DESCRIPTION
 *	_cos_sin14 computes the cosine of "angle" (expressed in normalized 
 *      radians) to an accuracy of at least 14 bits for angle of 0.0 to 
 *	1.0. Angles outside this range are less accurate. Normalized 
 *	radians mean that an angle of 1.0 corresponds to 2*PI radians. 
 *      The cosine is computed using an 8 segment piecewise, quadratic  
 *      approximation over the interval [0, 0.25) normalized radians.
 *	Symmetry of the cosine function is used to produce results over
 *	the remainder of the interval. The quadratic evaluation is of the form 
 *		A*x*x + B*x + C, 
 *	where x is the 1.0 plus the fractional input angle.
 */

/* Precomputed quadratic coefficients */
static float cos14_A[8] = {
  -19.62842, -18.87403, -17.39576,-15.24641,
  -12.51265,  -9.298583, -5.724892,-1.933483 };
static float cos14_B[8] = {
   39.25535,  37.69651, 34.55251,  29.84836,
   23.69532,  16.26119,  7.772535, -1.469602 };
static float cos14_C[8] = {
  -18.62693, -17.82171,-16.14998, -13.57606,
  -10.11380,  -5.815027,-0.774209,  4.858068 };


static __inline float _cos_sin14(float angle)
{
  unsigned int idx;
  unsigned int negflag;
  float work2;
  union {
    float f;
    unsigned int ui;
  } frac, work1, result;
  
  work1.f = angle;
  work1.ui &= 0x7ffffffe; 
  work2 = (float)((int)(work1.f));	/* work2 = INT(work1) */
  frac.f = 1.0f + (work1.f - work2);	/* frac = 1.0 + FRAC(work1) */

  negflag =  (frac.ui << 9)  & 0x80000000;
  if (frac.ui & 0x00200000) {
    /* between PI/2 and 3*PI/2 */
    frac.ui ^= 0x001fffff;
    negflag ^= 0x80000000;
  }
  frac.ui &= 0xff9fffff;
  idx = (frac.ui >> 18) & 0x7;
  result.f = cos14_A[idx] * frac.f + cos14_B[idx];
  result.f = result.f * frac.f + cos14_C[idx];
  result.ui ^= negflag;
  return (result.f);
}

#endif /* _COS_SIN14_H_ */
